## ------------------------------------------------------------------------
##  $Id: Abstract.pm,v 1.7 2002/12/16 06:45:07 joko Exp $
## ------------------------------------------------------------------------
##  $Log: Abstract.pm,v $
##  Revision 1.7  2002/12/16 06:45:07  joko
##  + injecting property 'guid' into each conrete class when initializing the schema
##
##  Revision 1.6  2002/11/17 07:21:47  jonen
##  + using Class::Tangram::import_schema() again for proper usage of its helper functions
##
##  Revision 1.5  2002/11/17 07:12:06  joko
##  + extended ::getProperties with argument "want_transactions"
##
##  Revision 1.4  2002/11/15 13:21:39  joko
##  + added capability to propagate global schema options from schema module to tangram schema
##
##  Revision 1.3  2002/11/09 01:27:23  joko
##  + dependency of Class::Tangram now is 1.13 at minimum
##
##  Revision 1.2  2002/10/25 11:45:10  joko
##  + enhanced robustness
##  + more logging for debug-levels
##
##  Revision 1.1  2002/10/17 00:07:14  joko
##  + new
## ------------------------------------------------------------------------
                       

package Data::Storage::Schema::Abstract;

use strict;
use warnings;
  
use Tangram;
use Class::Tangram;

use Tangram::RawDate;
use Tangram::RawTime;
use Tangram::RawDateTime;

use Tangram::FlatArray;
use Tangram::FlatHash;
use Tangram::PerlDump;


use Data::Dumper;

# get logger instance
my $logger = Log::Dispatch::Config->instance;



# ========   object constructor - %args to $self  ========
  sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my @args = ();
    @_ && (@args = @_);
    
    $logger->debug( __PACKAGE__ . "->new" );

    #my $self = { @_ };

    # merge all entries from all hashes in args-array to single array (key, val, key, val)
    my @args_all = ();
    map { push(@args_all, %{$_}); } @args;
    # make hash from array
    my %args = @args_all;

    #$logger->debug( __PACKAGE__ . "->new( @args_all )" );
    
    my $self = { %args };
    bless $self, $class;

    #use Data::Dumper; print Dumper($self);

    # remember as which Class we are instantiated
    $self->{invocant} = $invocant;

    $self->_checkClassTangram();
    $self->_init();
    
    return $self;
  }
  
  sub _checkClassTangram {
    my $self = shift;
    # check version of "Class::Tangram"
    my $ctversion = $Class::Tangram::VERSION;
    #print "ctversion: ", $ctversion, "\n";
    #if ($ctversion ne 1.04) {
    if ($ctversion lt 1.13) {
      print "Version of \"Class::Tangram\": ", $ctversion, "\n";
      print "minimum required version   : ", "1.13", "\n";
      #print "quitting!!! (please install older version (1.04) of \"Class::Tangram\"", "\n";
      exit;
    }
  }

  sub _parsePackageNamesFromFile {
    my $self = shift;
    my $filename = shift;
    my $exclude_pattern = shift;
    $logger->debug( __PACKAGE__ . "->_parsePackageFile( ... )" );
    if (!$filename) {
      $logger->error( __PACKAGE__ . "->_parsePackageFile: you must specify a filename" );
      return;
    }
    if (! -e $filename) {
      $logger->error( __PACKAGE__ . "->_parsePackageFile: could not find $filename" );
    }
    my @packages = ();
    open(FH, '<' . $filename);
    while (<FH>) {
      chomp();
      my $good = 1;
      $exclude_pattern && ($good = !m/$exclude_pattern/);
      push @packages, $1 if m/^package (.+?);/ && $good;
    }
    close(FH);
    return \@packages;
  }

  sub _doinit {
    my $self = shift;
    
    $logger->debug( __PACKAGE__ . "->_doinit()" );

    # WHAT?
    # imports classes to Tangram-Schema in this style, but in an abstracted way:
    #     Class::Tangram::import_schema("SystemEvent");
    # TODO:
    # move this code to Data::Storage::Schema::Tangram

    # TODO:
    # two cases:
    # 1. use a given list of strings as to-be-used-object-names
    # 2. parse the content of the given package and use all declared Perl Packages as objects
    
    my $classlist = $self->{EXPORT_OBJECTS};
    
    # parse complete package to build list of classes
    if ($#{$classlist} == -1) {
      $logger->debug( "_doinit() Initializing all Classes from " . $self->{invocant} );
      if (!$self->{invocant}->can('getFilename')) {
        $logger->error( __PACKAGE__ . "->_doinit: Package $self->{invocant} doesn't contain method 'getFilename'" );
        return;
      }
      my $perl_packagefile = eval('return ' . $self->{invocant} . '::getFilename();');
      my $packages = $self->_parsePackageNamesFromFile($perl_packagefile, $self->{invocant});
      # TODO:
      # cache list so this routine isn't called on every startup/request
      $classlist = $packages;
    }
    
    my $schema_properties = {};
    if ($self->{invocant}->can('getProperties')) {
      # TODO: rework this! call by ref and/or do oo
      #$schema_properties = eval('return ' . $self->{invocant} . '::getProperties();');
      $schema_properties = eval('return ' . $self->{invocant} . '::getProperties($self->{want_transactions});');
    }
    
    # build temporary index to look up class by classname
    my $classindex = {};
    
    # build argument to be passed to "Tangram::Schema->new(...)"
    my @classes;
    map {
      $logger->debug( __PACKAGE__ . "->_doinit: Initializing Class $_" );
      Class::Tangram::import_schema($_);
      # classname
      push @classes, $_;
      # schema
      my $schema;
      my $evalstr = '$schema = $' . $_ . '::schema;';
      my $res = eval($evalstr);
      push @classes, $schema;
      $classindex->{$_} = $schema;
    } @{$classlist};

    # build up the schema as a hash
    my $schema = { 
      classes => \@classes,
    };
    
    # merge schema properties
    foreach (keys %$schema_properties) {
      $schema->{$_} = $schema_properties->{$_};
    }

#print Dumper($schema);

    # patch guid into each concrete (non-abstract) object
    # TODO: refactor this code into function: _injectProperty('guid')
    foreach (@{$schema->{classes}}) {
      my $ref = ref $_;
      next if !$ref || $ref ne 'HASH';
      next if $_->{abstract};

#print Dumper($_);

      if (!$_->{fields}->{string}) {
        $_->{fields}->{string}->{guid} = undef;
        
      } elsif (ref $_->{fields}->{string} eq 'ARRAY') {
        push @{$_->{fields}->{string}}, 'guid';

      } elsif (ref $_->{fields}->{string} eq 'HASH') {
        $_->{fields}->{string}->{guid} = undef;
        
      }
    }
    
    # create a new Tangram schema object
    # a) these classes must occour in the same order used at a previous setup or
    # b) better use class ids!
    $self->{schema} = Tangram::Schema->new($schema);
    return 1;
    
=pod
    # ======================================
    # schema template
    $self->{schema} = Tangram::Schema->new({ 

        classes => [ 
          # misc
            'Person' => $Person::schema,
        ],

        make_object => sub { ... },
        set_id => sub { ... }
        get_id => sub { ... }
        normalize => sub { ... },

        control => '...'

        sql => { ... },

      });
    # ======================================
=cut

  }
  
  sub _init {
    my $self = shift;
    if (!$self->{schema}) {
      $self->_doinit();
    }
  }

  # return schema
  sub getSchema {
    my $self = shift;
    if (!$self->{schema}) {
      #print "schema is not yet initialized!", "\n";
      #exit;
      $logger->error( __PACKAGE__ . "->getSchema: Schema is not yet initialized!" );
      return;
    }
    return $self->{schema};
  }


1;
