## ------------------------------------------------------------------------
##  $Id: Trace.pm,v 1.1 2002/12/19 01:14:48 cvsjoko Exp $
## ------------------------------------------------------------------------
##  $Log: Trace.pm,v $
##  Revision 1.1  2002/12/19 01:14:48  cvsjoko
##  + initial check-in
##
## ------------------------------------------------------------------------


package BizWorks::Component::Trace;

use strict;
use warnings;

use Data::Dumper;

my $logger = Log::Dispatch::Config->instance;

sub __init {
  my $self = shift;
  my $action = shift;
  #my $options = shift;   # ??? or: $self->{options} already...?
  
  # set passed-in arguments
  $self->{action} = $action;

  # set default arguments
  $self->{classname} = ref $self;
  $self->{caller} = caller;
  
  # set default values
  $self->{action} ||= '';
  $self->{options}->{tag} ||= '';
  $self->{level} ||= 'info';
}

sub __handle_options {
  my $self = shift;
  #my $options = shift;   # ??? or: $self->{options} already...?
}

sub __get_prefix {
  my $self = shift;
  return __PACKAGE__ . "->$self->{action}: '$self->{caller}' (mixed into '$self->{classname}'): ($self->{options}->{tag}) - ";
}

sub _log {
  my $self = shift;
  my $message = shift;
  $self->{level} = shift;    # TODO: implement propagation of this into Log::Dispatch-object '$logger'!
  $self->{options} = shift;   # TODO: implement additional behaviours: e.g. { exit => 1 } would exit after tracedump
  
  $self->__init('_log');
  
  $logger->log( level => $self->{level}, message => $self->__get_prefix() . $message );
}

sub _skip {
  my $self = shift;
  my $message = shift;
  my $condition = shift;
  my $level = shift;    # TODO: implement propagation of this into Log::Dispatch-object '$logger'!
  my $options = shift;   # TODO: implement additional behaviours: e.g. { exit => 1 } would exit after tracedump

  my $classname = ref $self;
  my $caller = caller;

  $options->{tag} ||= '';

  if ($condition) {
    $logger->warning( __PACKAGE__ . "->_skip: '$caller' (mixed into '$classname'): ($options->{tag}) $message.");
    return 1;
  }
}

sub _trace {
  my $self = shift;
  my $data = shift;
  my $condition = shift;
  my $level = shift;    # TODO: implement propagation of this into Log::Dispatch-object '$logger'!
  my $options = shift;   # TODO: implement additional behaviours: e.g. { exit => 1 } would exit after tracedump
  
  my $classname = ref $self;
  my $caller = caller;

  $options->{tag} ||= '';

  # don't do _any_ tracing if disabled
  return if $self->{TRACE_DISABLED};
  
  # overwrite condition with local configured want-TRACE
  $condition ||= $self->{TRACE};
  
  # dump data
    my $result = 0;
    if ($condition) {
      my $message = Dumper($data);
      $logger->info( __PACKAGE__ . "->_trace: '$caller' (mixed into '$classname'): ($options->{tag}) $message.");
      $result = 1;
    } else {
      $result = 0;
    }

  # handle additional options
  if ($self->{TRACE_OPTIONS}) {
    exit if $options->{exit};
  }

  return $result;
}

1;
